;==================================================================================
; Switch-80 - 8080 mode utility - S170420
; HW ref: A250220 (V20-MBC)
;
; REQUIRED: IOS S260320 (or following revisions until otherwise stated)
;
; Assemble with "nasm -f bin  filename.asm -o filename.bin"
;
; Set the 8080 mode of the V20 CPU and execute the 8080 machine code already
; loaded into the RAM at the Physical address 0x1000:0000.
;
; The 8080 machine code must be already loaded into the RAM.
;
;==================================================================================
;
;
;  Physical memory layout (not in scale):
;
;
;  +-------------------------+
;  ! 0x0000:0x0000   0x00000 !    Switch-80 (local data area + program + reserved area)
;  !  ---------------------  !
;  ! 0x0000:0xFEFF   0x0FEFF !
;  +-------------------------+
;  ! 0x0FF0:0x0000   0x0FF00 !    Area available/used for the 8080 executable (64KB)
;  !  ---------------------  !    
;  ! 0x0FF0:0xFFFF   0x1FEFF !
;  +-------------------------+
;  ! 0x1000:0xFF00   0x1FF00 !    Not used - Reserved
;  !  ---------------------  !    (the V20 reset vector is here if RAM = 128KB)
;  ! 0x1000:0xFFFF   0x1FFFF !
;  +-------------------------+
;  ! 0x2000:0x0000   0x20000 !    Not used (can be not fitted with RAM)
;  !  ---------------------  !
;  ! 0xF000:0xFFFF   0xFFFFF !
;  +-------------------------+
; 
; The available Physical memory area for the 8080 user program is from 
; address 0x0FF0:0000 to 0x0FF0:0xFFFF (65.536 bytes).
;
; Switch-80 can work with the A250220 minimum RAM configuration (128KB).
; 

;
; NASM setting
;
    CPU     8086        ; Set 8086/8088 opcodes only
    BITS    16          ; Set default 16 bit

;
; Macro for the BRKEM instruction (V20 specific)
;
%macro brkem 1
  db 0fh
  db 0ffh
  db %1
%endmacro


;******************************************************************************
;******************************************************************************
;***
;*** Main program and data
;***
;******************************************************************************
;******************************************************************************

;------------------------------------------------------------------------------
;---
;---  Costants definitions (equates)
;---
;------------------------------------------------------------------------------


emu80_seg       equ    0x0FF0           ; 8080 RAM area segment (must match SEG_8080CODE of IOS)
loader_ram      equ    0                ; First RAM location used by iLoad-80
eos             equ    0x00             ; End of string
cr              equ    0x0d             ; Carriage return
lf              equ    0x0a             ; Line feed
space           equ    0x20             ; Space
rx_port         equ    0x01             ; IOS serial Rx read port
opcode_port     equ    0x01             ; IOS opcode write port
exec_wport      equ    0x00             ; IOS execute opcode write port
tx_opcode       equ    0x01             ; IOS serial Tx operation opcode

;------------------------------------------------------------------------------
;---
;---  Switch-80 IRQ vectors area
;---
;------------------------------------------------------------------------------

                org    loader_ram
page_0:
    jmp     starting_addr               ; Vector 0 to 31 reserved area
    times   128-$+page_0    db      0
    
vector_32_off:                          ; Vector 32: jump to the 8080 code
    dw      0
vector_32_seg:
    dw      emu80_seg
    
vector_33_off:                          ; Vector 33: Serial Rx ISR
    dw      vect33
vector_33_seg:
    dw      0

;------------------------------------------------------------------------------
;---
;--- Vector 33 ISR and equates (Serial Rx)
;---
;------------------------------------------------------------------------------

;
; Vector 33 ISR equates
;
; NOTE:
; These equates must be the same in the user program (8080 code) to
; enable the use of the serial Rx IRQ and ISR (if used).
SER_BUFSIZE     EQU     0x3F
SER_FULLSIZE    EQU     0x30
serBuf          EQU     0x2000
serRdPtr        EQU     serInPtr+2
serBufUsed      EQU     serRdPtr+2
serInPtr        EQU     serBuf+SER_BUFSIZE

;
; Vector 33 ISR
;
vect33:
    push    ax
    push    bx
    in      al, rx_port     ; Read a char from serial
    mov     ah, al          ; Save it in AH
    mov     al, [serBufUsed]
    cmp     al, SER_BUFSIZE ; If full then ignore
    jnz     notFull
    mov     al, ah          ; Char in AL again
    jmp     rts0
notFull:
    mov     bx,[serInPtr]
    inc     bx
    mov     al, bl          ; Only need to check low byte becasuse buffer<256 bytes
    cmp     al, ((serBuf+SER_BUFSIZE) & 0xff)
    jnz     notWrap
    mov     bx, serBuf
notWrap:
    mov     [serInPtr], bx
    mov     al, ah          ; Char in AL again
    mov     [bx], al
    mov     al, [serBufUsed]
    inc     al
    mov     [serBufUsed], al
    cmp     al, SER_FULLSIZE
    jc      rts0
rts0:           
    pop     bx
    pop     ax
    sti                     ; Re-enable interrupts
    iret

;------------------------------------------------------------------------------
;---
;---  Local stack pointer and data area
;---
;------------------------------------------------------------------------------

start:
    jmp     starting_addr           ; Inside the stack area, can be overwritten
strAddr:
    dw      0                       ; Starting addres of the 8080 code.
                                    ; The addres must be equal to SWITCH80_STRADDR (IOS) 
                                    ; NOTE: this word is updated dinamically at
                                    ; load time by the IOS with the real value
binSize:
    dw      0                       ; Size in bytes of the 8080 code.
                                    ; The addres must be equal to SWITCH80_STRADDR (IOS)
                                    ; NOTE: this word is updated dinamically at
                                    ; load time by the IOS with the real value
                                    ; A 0 value means 65536 bytes (see IOS)
stackArea:
    times   32-$+stackArea  db  0   ; Local stack area
local_stack:

;------------------------------------------------------------------------------
;---
;---  Main program
;---
;------------------------------------------------------------------------------

;
; Init
;
starting_addr:
    mov     sp, local_stack ; Set the stack
    mov     ax, cs          ; DS = SS = CS
    mov     ds, ax
    mov     ss, ax
    mov     ax, emu80_seg   ; ES = emu80_seg (is the segment used to load the 8080 code)
    mov     es, ax
;
; Print a welcome message
;
    call    crlf
    mov     bx, hello_msg
    call    puts
    call    crlf
    call    crlf
;
; Test for NEC V20
;
    mov     ax,0x0101
    aad     0x10            ; NEC V20 ignores the argument
    cmp     al,0x0B         ; and always does AL = AL * 0x0A + AH
    je      isNecV20        ; Jump if NEC V20 found
;
; Not a NEC V20, print a message and halt
;
    mov     bx, ih_load_msg_4
    call    puts
    mov     bx, load_msg_4
    call    puts
    mov     bx, sad_msg
    call    puts
    hlt
;
; Set the offset (starting address) for Vector 32 (jump to the 8080 area) 
; and print the starting address
;
isNecV20:
    mov     bx, ih_load_msg_4
    call    puts
    mov     bx, load_msg_1
    call    puts
    mov     bx, es          ; Print the ES segment (8080 area segment)
    call    print_word
    mov     al, ':'
    call    putc
    mov     bx, [strAddr]   
    mov     [vector_32_off], bx ; Set the offset for Vector 32
    call    print_word      ; Print the starting addresss of the 8080 code
;
; Print the size of the 8080 code
;
    mov     bx, load_msg_5
    call    puts
    mov     ax, [binSize]
    or      ax, ax
    jnz     lessThenFFFF    ; Jump is size not 0
    mov     al, '1'         ; Size is 0, so it is 65536 (see IOS)
    call    putc            ; Print a leading 1
lessThenFFFF:
    mov     bx, [binSize]
    call    print_word      ; Print the size
    call    crlf
;
; Print a message ("Switching to 8080 mode...")
;
    mov     bx, ih_load_msg_4
    call    puts
    mov     bx, load_msg_3
    call    puts
    call    crlf
;
; Jump to the 8080 loaded code
;               
    mov     ax, es          ; DS = 8080 RAM area segment 
    mov     ds, ax
    brkem   32              ; Enable 8080 mode and jump at emu80_seg:strAddr
halt:
    hlt                     ; Just if there is a return from the 8080 mode
    jmp     halt

;------------------------------------------------------------------------------
;---
;---  Message definitions
;---
;------------------------------------------------------------------------------

hello_msg       db   "Switch-80 - 8080 mode for V20 CPU - S170420", eos
load_msg_1      db   "8080 code Starting Address: ", eos
load_msg_3      db   "Switching to 8080 mode...", cr, lf, eos
load_msg_4      db   "You are not using a V20 CPU! - System halted. Sorry...", cr, lf, eos
load_msg_5      db   " - Size: ", eos
ih_load_msg_4   db   "Switch-80: ", eos
sad_msg         db   cr, lf
                db   "     .-''''''-.    ", cr, lf
                db   "   .'          '.  ", cr, lf
                db   "  /   O      O   \ ", cr, lf
                db   " :           `    :", cr, lf
                db   " |                |", cr, lf
                db   " :    .------.    :", cr, lf
                db   "  \  '        '  / ", cr, lf
                db   "   '.          .'  ", cr, lf
                db   "     '-......-'    ", cr, lf, lf, eos


;******************************************************************************
;******************************************************************************
;***
;*** Subroutines
;***
;******************************************************************************
;******************************************************************************

;------------------------------------------------------------------------------
;---
;--- String subroutines
;---
;------------------------------------------------------------------------------

;
; Send a string to the serial line, BX contains the pointer to the string:
;
puts:
    push    ax
    push    bx
puts_loop:
    mov     al, [bx]
    cmp     al, eos         ; End of string reached?
    jz short      puts_end      ; Yes
    call    putc
    inc     bx              ; Increment character pointer
    jmp     puts_loop       ; Transmit next character
puts_end:
    pop     bx
    pop     ax
    ret

;
; Print_word prints the four hex digits of a word to the serial line. The 
; word is expected to be in BX.
;
print_word:
    push    bx
    push    ax
    mov     al, bh          ; Print high digits
    call    print_byte
    mov     al, bl          ; Print low digits
    call    print_byte
    pop     ax
    pop     bx
    ret

;
; Print_byte prints a single byte in hexadecimal notation to the serial line.
; The byte to be printed is expected to be in AL.
;
print_byte:
    push    ax              ; Save the contents of the registers
    push    cx
    mov     ah, al
    mov     cl, 4
    ror     al, cl
    call    print_nibble    ; Print high nibble
    mov     al, ah
    call    print_nibble    ; Print low nibble
    pop     cx              ; Restore original register contents
    pop     ax
    ret

;
; Print_nibble prints a single hex nibble which is contained in the lower 
; four bits of AL:
;
print_nibble:    
    push    ax              ; We won't destroy the contents of A
    and     al, 0xf         ; Just in case...
    add     al, '0'         ; If we have a digit we are done here.
    cmp     al, '9' + 1     ; Is the result > 9?
    jc      print_nibble_1
    add     al, 'A' - '0' - 0xa  ; Take care of A-F
print_nibble_1:  
    call    putc            ; Print the nibble and
    pop     ax              ; restore the original value of A
    ret

;
; Send a CR/LF pair:
;
crlf:           
    push    ax
    mov     al, cr
    call    putc
    mov     al, lf
    call    putc
    pop     ax
    ret

;------------------------------------------------------------------------------
;---
;--- I/O suroutines
;---
;------------------------------------------------------------------------------

;
; Send a single character to the serial line (AL contains the character):
;
putc:
    mov     ah, al          ; Save the chat into AH
    mov     al, tx_opcode   ; AL = IOS Serial Tx operation opcode
    out     opcode_port, al ; Send to IOS the Tx operation opcode
    mov     al, ah          ; Output char into AL
    out     exec_wport, al  ; Write AL to the serial
    ret
        