;==================================================================================
; iLoad-80 - Intel-Hex Loader - S230220
; HW ref: A250220 (V20-MBC)
;
; REQUIRED: IOS S260320 (or following revisions until otherwise stated)
;
; Assemble with "nasm -f bin  filename.asm -o filename.bin"
;
; Loads an 8080 executable in Intel-Hex format, sets the V20 8080 mode and executes it. 
; The executable must be an 8080 machine code "flat file" inside a single segment,
; so the maximum executable size is 64KB (that is also the maximum address space allowed
; with the used Intel-Hex format.
;
; The 8080 executable file is loaded and executed into the segment at 0x0FF0.
;
; The starting address of the loaded executable is 0x0FF0:STRADDR, where STRADDR is the
; first address (16 bit) of the Intel-Hex stream (the first executed instruction starts 
; at the address of the first byte of the Intel-Hex stream).
;
; iLoad-80 checks for errors in the Intel-Hex stream.
;
;==================================================================================
;
;
;  Physical memory layout (not in scale):
;
;  +-------------------------+
;  ! 0x0000:0x0000   0x00000 !    iLoad-80 (local data area + program + reserved area)
;  !  ---------------------  !
;  ! 0x0000:0xFEFF   0x0FEFF !
;  +-------------------------+
;  ! 0x0FF0:0x0000   0x0FF00 !    Area available/used for the 8080 executable (64KB)
;  !  ---------------------  !    
;  ! 0x0FF0:0xFFFF   0x1FEFF !
;  +-------------------------+
;  ! 0x1000:0xFF00   0x1FF00 !    Not used - Reserved
;  !  ---------------------  !    (the V20 reset vector is here if RAM = 128KB)
;  ! 0x1000:0xFFFF   0x1FFFF !
;  +-------------------------+
;  ! 0x2000:0x0000   0x20000 !    Not used (can be not fitted with RAM)
;  !  ---------------------  !
;  ! 0xF000:0xFFFF   0xFFFFF !
;  +-------------------------+
; 
; The available Physical memory area for the 8080 user program is from 
; address 0x0FF0:0000 to 0x0FF0:0xFFFF (65.536 bytes).
;
; iLoad-80 can work with the A250220 minimum RAM configuration (128KB).
; 
; Multi-segment executables are not supportd by iLoad-80.
; 
;==================================================================================
;
; NOTE: I've ported to the 8086/8088 assembler the original iLoad (SW ref. S200718)
;       for the Z80-MBC2, translating the source line by line.
;       In the following the "convention" that I've (mainly) used for the registers.
;
;       16 bit registes (Z80 -> 8086):
;               AF -> AX
;               BC -> CX
;               DE -> DX
;               HL -> BX
;
;       8 bit registes (Z80 -> 8086):
;                A -> AL
;                B -> CH
;                C -> CL
;                D -> DH
;                E -> DL
;                H -> BH
;                L -> BL
;
;==================================================================================

;
; NASM setting
;
    CPU     8086        ; Set 8086/8088 opcodes only
    BITS    16          ; Set default 16 bit

;
; Macro for the BRKEM instruction (V20 specific)
;
%macro brkem 1
  db 0fh
  db 0ffh
  db %1
%endmacro


;******************************************************************************
;******************************************************************************
;***
;*** Main program and data
;***
;******************************************************************************
;******************************************************************************

;------------------------------------------------------------------------------
;---
;---  Costants definitions (equates)
;---
;------------------------------------------------------------------------------


emu80_seg       equ    0x0ff0           ; 8080 RAM area segment (must match SEG_8080CODE of IOS)
loader_ram      equ    0                ; First RAM location used by iLoad-80
eos             equ    0x00             ; End of string
cr              equ    0x0d             ; Carriage return
lf              equ    0x0a             ; Line feed
space           equ    0x20             ; Space
rx_port         equ    0x01             ; IOS serial Rx read port
opcode_port     equ    0x01             ; IOS opcode write port
exec_wport      equ    0x00             ; IOS execute opcode write port
tx_opcode       equ    0x01             ; IOS serial Tx operation opcode

;------------------------------------------------------------------------------
;---
;---  iLoad-80 IRQ vectors area
;---
;------------------------------------------------------------------------------

                org    loader_ram
page_0:
    jmp     starting_addr               ; Vector 0 to 31 reserved area
    times   128-$+page_0    db      0
    
vector_32_off:                          ; Vector 32: jump to the 8080 code
    dw      0
vector_32_seg:
    dw      emu80_seg
    
vector_33_off:                          ; Vector 33: Serial Rx ISR
    dw      vect33
vector_33_seg:
    dw      0

;------------------------------------------------------------------------------
;---
;--- Vector 33 ISR and equates (Serial Rx)
;---
;------------------------------------------------------------------------------

;
; Vector 33 ISR equates
;
; NOTE:
; These equates must be the same in the user program (8080 code) to
; enable the use of the serial Rx IRQ and ISR (if used).
SER_BUFSIZE     EQU     0x3F
SER_FULLSIZE    EQU     0x30
serBuf          EQU     0x2000
serRdPtr        EQU     serInPtr+2
serBufUsed      EQU     serRdPtr+2
serInPtr        EQU     serBuf+SER_BUFSIZE

;
; Vector 33 ISR
;
vect33:
    push    ax
    push    bx
    in      al, rx_port     ; Read a char from serial
    mov     ah, al          ; Save it in AH
    mov     al, [serBufUsed]
    cmp     al, SER_BUFSIZE ; If full then ignore
    jnz     notFull
    mov     al, ah          ; Char in AL again
    jmp     rts0
notFull:
    mov     bx,[serInPtr]
    inc     bx
    mov     al, bl          ; Only need to check low byte becasuse buffer<256 bytes
    cmp     al, ((serBuf+SER_BUFSIZE) & 0xff)
    jnz     notWrap
    mov     bx, serBuf
notWrap:
    mov     [serInPtr], bx
    mov     al, ah          ; Char in AL again
    mov     [bx], al
    mov     al, [serBufUsed]
    inc     al
    mov     [serBufUsed], al
    cmp     al, SER_FULLSIZE
    jc      rts0
rts0:           
    pop     bx
    pop     ax
    sti                     ; Re-enable interrupts
    iret

;------------------------------------------------------------------------------
;---
;---  Stack pointer local area
;---
;------------------------------------------------------------------------------

start:
    jmp     starting_addr           ; Inside the stack area, can be overwritten
    times   32-$+start  db      0   ; Local stack area
local_stack:

;------------------------------------------------------------------------------
;---
;---  Main program
;---
;------------------------------------------------------------------------------

;
; Init
;
starting_addr:
    mov     sp, local_stack ; Set the stack
    mov     ax, cs          ; DS = SS = CS
    mov     ds, ax
    mov     ss, ax
    mov     ax, emu80_seg   ; ES = emu80_seg
    mov     es, ax
;
; Print a welcome message
;
    mov     bx, hello_msg
    call    puts
    call    crlf
;
; Test for NEC V20
;
    mov   ax,0x0101
    aad   0x10              ; NEC V20 ignores the argument
    cmp   al,0x0B           ; and always does AL = AL * 0x0A + AH
    je    isNecV20          ; Jump if NEC V20 found
;
; Not a NEC V20, print a message and halt
;
    call    crlf
    mov     bx, ih_load_msg_4
    call    puts
    mov     bx, load_msg_4
    call    puts
    mov     bx, sad_msg
    call    puts
    hlt
;
; Load an INTEL-Hex file into memory
;
isNecV20:
    call    ih_load         ; Load Intel-Hex file
    mov     al, 0xff        ; Test for errors
    cmp     al, bh          ; (A = H ?)
    jnz     short print_addr; Jump if BH (H) <> $FF (no errors)
    cmp     al, bl          ; (A = L ?)
    jnz     short print_addr; Jump if BL (L) <> $FF (no errors)
;
; Print an error message and halt cpu
;               
    mov     bx, ih_load_msg_4
    call    puts
    mov     bx, load_msg_2
    call    puts
    hlt
;
; Print starting address
;
print_addr:
    push    bx              ; Save starting addresss BX (HL)
    mov     bx, ih_load_msg_4
    call    puts
    mov     bx, load_msg_1
    call    puts
    mov     bx, es          ; Print the ES segment (8080 area segment)
    call    print_word
    mov     al, ':'
    call    putc
    pop     bx              ; Print the starting addresss of the loaded program
    call    print_word
    call    crlf
    push    bx              ; Print a message about the 8080 mode activation
    mov     bx, ih_load_msg_4
    call    puts
    mov     bx, load_msg_3  
    call    puts
    call    crlf
    pop     bx
;
; Flush remaining input data (if any) and jump to the loaded program
;               
flush_rx:
    in      al, rx_port     ; Read a char from serial port
    cmp     al, 0xff        ; Is <> $FF?
    jnz     short flush_rx  ; Yes, read an other one
    mov     [vector_32_off], bx; Store the 8080 starting addr (offset) into the vector 32 offset
    mov     ax, es          ; DS = 8080 RAM area segment 
    mov     ds, ax
    brkem   32              ; Enable 8080 mode and jump
halt:
    hlt                     ; Just if there is a return from the 8080 mode
    jmp     halt

;------------------------------------------------------------------------------
;---
;---  Message definitions
;---
;------------------------------------------------------------------------------

hello_msg       db   "iLoad-80 - Intel-Hex Loader (V20 CPU) - S230220", eos
load_msg_1      db   "Starting Address: ", eos
load_msg_2      db   "Load error - System halted", eos
load_msg_3      db   "Switching to 8080 mode...", cr, lf, eos
load_msg_4      db   "You are not using a V20 CPU! - System halted. Sorry...", cr, lf, eos
ih_load_msg_1   db   "Waiting input stream...", eos
ih_load_msg_2   db   "Syntax error!", eos
ih_load_msg_3   db   "Checksum error!", eos 
ih_load_msg_4   db   "iLoad-80: ", eos
sad_msg         db   cr, lf
                db   "     .-''''''-.    ", cr, lf
                db   "   .'          '.  ", cr, lf
                db   "  /   O      O   \ ", cr, lf
                db   " :           `    :", cr, lf
                db   " |                |", cr, lf
                db   " :    .------.    :", cr, lf
                db   "  \  '        '  / ", cr, lf
                db   "   '.          .'  ", cr, lf
                db   "     '-......-'    ", cr, lf, lf, eos


;******************************************************************************
;******************************************************************************
;***
;*** Subroutines
;***
;******************************************************************************
;******************************************************************************


;
; Load an INTEL-Hex file (a ROM image) into memory. This routine has been 
; more or less stolen from a boot program written by Andrew Lynch and adapted
; to this simple Z80 based machine.
;
; The first address in the INTEL-Hex file is considerd as the Program Starting Address
; and is stored into HL.
;
; If an error is found HL=$FFFF on return.
;
; The INTEL-Hex format looks a bit awkward - a single line contains these 
; parts:
; ':', Record length (2 hex characters), load address field (4 hex characters),
; record type field (2 characters), data field (2 * n hex characters),
; checksum field. Valid record types are 0 (data) and 1 (end of file).
;
; Please note that this routine will not echo what it read from stdin but
; what it "understood". :-)
; 
ih_load:
    push    ax              ; (AF)
    push    dx              ; (DE)
    push    cx              ; (BC)
    mov     cx, 0xffff      ; (Init BC = $FFFF)
    mov     bx, ih_load_msg_1
    call    puts
    call    crlf
ih_load_loop:
    call    getc            ; Get a single character
    cmp     al, cr          ; Don't care about CR
    jz      short ih_load_loop
    cmp     al, lf          ; ...or LF
    jz      short ih_load_loop
    cmp     al, space       ; ...or a space
    jz      short ih_load_loop
    call    to_upper        ; Convert to upper case
    call    putc            ; Echo character
    cmp     al, ':'         ; Is it a colon? 
    jnz     short ih_load_err   ; No - print an error message
    call    get_byte        ; Get record length into A
    mov     dh, al          ; Length is now in D (D = A)
    mov     dl, 0           ; Clear checksum
    call    ih_load_chk     ; Compute checksum
    call    get_word        ; Get load address into BX (HL)
    mov     al, 0xff        ; Save first address as the starting addr
    cmp     al, ch          ; (A = B ?)
    jnz     short update_chk    ; Jump if B<>$FF
    cmp     al, cl          ; (A = C ?)
    jnz     short update_chk; (Jump if C<>$FF)
    mov     cx, bx          ; Save starting address in CX (BC = HL)
update_chk:
    mov     al, bh          ; Update checksum by this address (A = H)
    call    ih_load_chk
    mov     al, bl          ; (A = L)
    call    ih_load_chk
    call    get_byte        ; Get the record type
    call    ih_load_chk     ; Update checksum
    cmp     al, 1           ; Have we reached the EOF marker?
    jnz     short ih_load_data  ; No - get some data
    call    get_byte        ; Yes - EOF, read checksum data
    call    ih_load_chk     ; Update our own checksum
    mov     al, dl          ; (A = E)
    and     al, al          ; Is our checksum zero (as expected)?
    jz      short ih_load_exit  ; Yes - exit this routine
ih_load_chk_err:
    call    crlf            ; No - print an error message
    mov     bx, ih_load_msg_4
    call    puts
    mov     bx, ih_load_msg_3
    call    puts
    mov     cx, 0xffff      ; (BC = 0xffff)
    jmp     ih_load_exit    ; ...and exit
ih_load_err:
    call    crlf
    mov     bx, ih_load_msg_4
    call    puts            ; Print error message
    mov     bx, ih_load_msg_2
    call    puts
    mov     cx, 0xffff      ; (BC = 0xffff)
ih_load_data:
    mov     al, dh          ; Record length is now in AL
    and     al, al          ; Did we process all bytes?
    jz      short ih_load_eol   ; Yes - process end of line
    call    get_byte        ; Read two hex digits into A
    call    ih_load_chk     ; Update checksum
store_byte:
    mov     [es:bx], al     ; Store byte into the 8080 RAM segment
    inc     bx              ; Increment pointer
    dec     dh              ; Decrement remaining record length
    jmp     ih_load_data    ; Get next byte
ih_load_eol:
    call    get_byte        ; Read the last byte in the line
    call    ih_load_chk     ; Update checksum
    mov     al, dl
    and     al, al          ; Is the checksum zero (as expected)?
    jnz     short ih_load_chk_err
    call    crlf
    jmp     ih_load_loop    ; Yes - read next line
ih_load_exit:
    call    crlf
    mov     bx, cx          ; BX = CX (HL = BC)
    pop     cx              ; Restore registers (BC)
    pop     dx              ; (DE)
    pop     ax              ; (AF)
    ret
;
; Compute DL = DL - AL (E = E - A)
;
ih_load_chk:
    sub     dl, al          ; DL = DL - AL
    ret

;------------------------------------------------------------------------------
;---
;--- String subroutines
;---
;------------------------------------------------------------------------------

;
; Get a word (16 bit) in hexadecimal notation. The result is returned in BX.
; Since the routines get_byte and therefore get_nibble are called, only valid
; characters (0-9a-f) are accepted.
;
get_word:
    push    ax
    call    get_byte        ; Get the upper byte
    mov     bh, al
    call    get_byte        ; Get the lower byte
    mov     bl, al
    pop     ax
    ret

;
; Get a byte in hexadecimal notation. The result is returned in AL. Since
; the routine get_nibble is used only valid characters are accepted - the 
; input routine only accepts characters 0-9, a-f.
;
get_byte:
    push    cx              ; Save CX
    call    get_nibble      ; Get upper nibble
    mov     cl, 4
    rol     al, cl
    mov     ah, al          ; Save upper four bits
    call    get_nibble      ; Get lower nibble
    or      al, ah          ; Combine both nibbles
    pop     cx              ; Restore CX
    ret

;
; Get a hexadecimal digit from the serial line. This routine blocks until
; a valid character (0-9a-f) has been entered. A valid digit will be echoed
; to the serial line interface. The lower 4 bits of AL contain the value of 
; that particular digit.
;
get_nibble:
    call    getc            ; Read a character in AL
    call    to_upper        ; Convert to upper case
    call    is_hex          ; Was it a hex digit?
    jnc     get_nibble      ; No, get another character
    call    nibble2val      ; Convert nibble to value
    call    print_nibble
    ret

;
; is_hex checks a character stored in AL for being a valid hexadecimal digit.
; A valid hexadecimal digit is denoted by a set C flag.
;
is_hex:
    cmp     al, 'F' + 1     ; Greater than 'F'?
    jc      is_hex_0        ; ret nc ; Yes return
    ret
is_hex_0:
    cmp     al, '0'         ; Less than '0'?
    jnc     is_hex_1        ; No, continue
    clc                     ; Clear carry
    ret
is_hex_1:
    cmp     al, '9' + 1     ; Less or equal '9*?
    jnc     is_hex_1b       ; ret c ; Yes
    ret
is_hex_1b:
    cmp     al, 'A'         ; Less than 'A'?
    jnc     is_hex_2        ; No, continue
    clc                     ; Yes - clear carry and return
    ret
is_hex_2:        
    stc                     ; Set carry
    ret

;
; Convert a single character contained in AL to upper case:
;
to_upper:
    cmp     al, 'a'             ; Nothing to do if not lower case
    jnc     to_upper1           ; ret     c
    ret
to_upper1:
    cmp     al, 'z' + 1         ; > 'z'?
    jc      to_upper2           ; ret nc ; Nothing to do, either
    ret
to_upper2:
    and     al, 0x5f             ; Convert to upper case
    ret

;
; Expects a hexadecimal digit (upper case!) in AL and returns the
; corresponding value in AL.
;
nibble2val:
    cmp     al, '9' + 1         ; Is it a digit (less or equal '9')?
    jc      nibble2val_1        ; Yes
    sub     al, 7               ; Adjust for A-F
nibble2val_1:
    sub     al, '0'             ; Fold back to 0..15
    and     al, 0xf             ; Only return lower 4 bits
    ret

;
; Send a string to the serial line, BX contains the pointer to the string:
;
puts:
    push    ax
    push    bx
puts_loop:
    mov     al, [bx]
    cmp     al, eos         ; End of string reached?
    jz short      puts_end      ; Yes
    call    putc
    inc     bx              ; Increment character pointer
    jmp     puts_loop       ; Transmit next character
puts_end:
    pop     bx
    pop     ax
    ret

;
; Print_word prints the four hex digits of a word to the serial line. The 
; word is expected to be in BX.
;
print_word:
    push    bx
    push    ax
    mov     al, bh          ; Print high digits
    call    print_byte
    mov     al, bl          ; Print low digits
    call    print_byte
    pop     ax
    pop     bx
    ret

;
; Print_byte prints a single byte in hexadecimal notation to the serial line.
; The byte to be printed is expected to be in AL.
;
print_byte:
    push    ax              ; Save the contents of the registers
    push    cx
    mov     ah, al
    mov     cl, 4
    ror     al, cl
    call    print_nibble    ; Print high nibble
    mov     al, ah
    call    print_nibble    ; Print low nibble
    pop     cx              ; Restore original register contents
    pop     ax
    ret

;
; Print_nibble prints a single hex nibble which is contained in the lower 
; four bits of AL:
;
print_nibble:    
    push    ax              ; We won't destroy the contents of A
    and     al, 0xf         ; Just in case...
    add     al, '0'         ; If we have a digit we are done here.
    cmp     al, '9' + 1     ; Is the result > 9?
    jc      print_nibble_1
    add     al, 'A' - '0' - 0xa  ; Take care of A-F
print_nibble_1:  
    call    putc            ; Print the nibble and
    pop     ax              ; restore the original value of A
    ret

;
; Send a CR/LF pair:
;
crlf:           
    push    ax
    mov     al, cr
    call    putc
    mov     al, lf
    call    putc
    pop     ax
    ret

;------------------------------------------------------------------------------
;---
;--- I/O suroutines
;---
;------------------------------------------------------------------------------

;
; Send a single character to the serial line (AL contains the character):
;
putc:
    mov     ah, al          ; Save the chat into AH
    mov     al, tx_opcode   ; AL = IOS Serial Tx operation opcode
    out     opcode_port, al ; Send to IOS the Tx operation opcode
    mov     al, ah          ; Output char into AL
    out     exec_wport, al  ; Write AL to the serial
    ret
    
;
; Wait for a single incoming character on the serial line
; and read it, result is in AL:
;
getc:           
    in      al, rx_port     ; Read a char from serial
    cmp     al, 0xff        ; It is = $FF?
    jz      short getc      ; If yes jump until a valid char is received
    ret
    